import datetime
import os

from petrus._core import utils
from petrus._core.calcs.Block import Block
from petrus._core.calcs.Calc import Calc
from petrus._core.calcs.Draft import Draft
from petrus._core.calcs.File import File
from petrus._core.calcs.Git import Git
from petrus._core.calcs.Project import Project
from petrus._core.calcs.Text import Text
from petrus._core.TOML import TOML


class Prog(Calc):
    _CORE = "kwargs"
    INPUTS = {
        "author": None,
        "description": None,
        "email": "Email of the author.",
        "requires_python": None,
        "github": None,
        "v": "Version string for the project.",
        "year": None,
    }

    def __post_init__(self):
        self.git.init()
        if self.git.is_repo():
            self.save("gitignore")
        self.pp["project"] = self.project.to_dict()
        print(self.pp["project"])
        self.pp["build-system"] = self.build_system
        self.pp[()] = utils.easy_dict(self.pp[()])
        self.text.pp = str(self.pp)
        self.save("license")
        self.save("manifest")
        self.save("pp")
        self.save("readme")
        self.save("setup")
        utils.run_isort()
        utils.run_black(os.getcwd())
        self.git.commit_version()
        self.git.push()
        utils.pypi()

    def _calc_author(self):
        f = lambda z: str(z).strip()
        n = f(self.kwargs["author"])
        e = f(self.kwargs["email"])
        x = n, e
        authors = self.project.authors
        if type(authors) is not list:
            return x
        for a in authors:
            if type(a) is not dict:
                continue
            n = f(a.get("name", ""))
            e = f(a.get("email", ""))
            y = n, e
            if y != ("", ""):
                return y
        return x

    def _calc_block(self):
        return Block(self)

    def _calc_build_system(self):
        ans = self.pp["build-system"]
        if type(ans) is dict:
            ans = utils.easy_dict(ans)
        if ans is not None:
            return ans
        ans = dict()
        ans["requires"] = ["setuptools>=61.0.0"]
        ans["build-backend"] = "setuptools.build_meta"
        ans = utils.easy_dict(ans)
        return ans

    def _calc_draft(self):
        return Draft(self)

    def _calc_file(self):
        return File(self)

    def _calc_git(self):
        return Git(self)

    def _calc_github(self):
        u = self.kwargs["github"]
        if u == "":
            return ""
        return f"https://github.com/{u}/{self.project.name}"

    def _calc_packages(self):
        utils.mkdir("src")
        ans = []
        for x in os.listdir("src"):
            if self._is_pkg(x):
                ans.append(x)
        if len(ans):
            return ans
        for x in os.listdir():
            if self._is_pkg(x):
                ans.append(x)
        if not self.file.exists("pp"):
            ans.append(self.project.name)
            if not self._is_pkg(self.project.name):
                self.save("init")
                self.save("main")
        ans = utils.easy_list(ans)
        return ans

    def _calc_pp(self):
        return TOML(text=self.text.pp)

    def _calc_project(self):
        return Project(self)

    def _calc_text(self):
        return Text(self)

    def _calc_year(self):
        return self.kwargs["year"] or str(datetime.datetime.now().year)

    @staticmethod
    def _is_pkg(path):
        if not os.path.isdir(path):
            return False
        f = os.path.join(path, "__init__.py")
        return utils.isfile(f)

    def save(self, n, /):
        file = getattr(self.file, n)
        text = getattr(self.text, n)
        root = os.path.dirname(file)
        if root and not os.path.exists(root):
            os.mkdir(root)
        with open(file, "w") as s:
            s.write(text)



import importlib.metadata
import importlib.resources
import os
import shutil
import string
import subprocess
import sys

import black
import isort
import requests

from petrus._core import utils


def dict_match(a, b, /):
    a = dict(a)
    b = dict(b)
    keys = set(a.keys()) & set(b.keys())
    ans = all(a[k] == b[k] for k in keys)
    return ans


def easy_list(iterable):
    ans = sorted(list(set(iterable)))
    ans = list(filter(lambda x: x, ans))
    return ans


def easy_dict(dictionary, *, purge=False):
    d = dict(dictionary)
    keys = sorted(list(d.keys()))
    keys = filter(lambda k: d[k], keys)
    ans = {k: d[k] for k in keys}
    return ans


def fix_dependency(line, /):
    dependency = line.strip()
    chars = set(dependency)
    chars -= set(string.ascii_letters)
    chars -= set(string.digits)
    chars -= set("-_")
    if len(chars):
        return dependency
    version = _get_some_version(dependency)
    if version is None:
        return dependency
    dependency += ">=" + version
    return dependency


def run_black(path):
    try:
        return black.main([path])
    except:
        pass


def run_isort():
    files = []
    walk = os.walk(os.getcwd())
    for root, dnames, fnames in walk:
        for fname in fnames:
            f = os.path.join(root, fname)
            files.append(f)
    for f in files:
        if os.path.splitext(f)[1] == ".py":
            isort.file(f)


def isdir(path):
    if not os.path.exists(path):
        return False
    if not os.path.isdir(path):
        raise ValueError
    return True


def isfile(path):
    if not os.path.exists(path):
        return False
    if not os.path.isfile(path):
        raise ValueError
    return True


def mkdir(path):
    if isdir(path):
        return
    os.mkdir(path)


def py(*args):
    args = [sys.executable, "-m"] + list(args)
    return subprocess.run(args)


def pypi():
    shutil.rmtree("dist", ignore_errors=True)
    if py("build").returncode:
        return
    subprocess.run(["twine", "upload", "dist/*"])


def walk(path, *, recursively):
    if not os.path.exists(path):
        return (x for x in ())
    if not recursively:
        ans = os.listdir(path)
        ans = (os.path.join(path, n) for n in ans)
        ans = filter(os.path.isfile, ans)
        for f in ans:
            yield f
        return
    for root, dnames, fnames in os.walk(path):
        for fname in fnames:
            yield os.path.join(root, fname)


def _get_some_version(pkg, /):
    return _get_local_version(pkg) or _get_latest_version(pkg)


def _get_local_version(pkg, /):
    try:
        ans = importlib.metadata.version(pkg)
    except:
        return None
    url = "https://pypi.org/pypi/%s/%s" % (pkg, ans)
    r = requests.get(url)
    if r.status_code == 404:
        return None
    return ans


def _get_latest_version(pkg, /):
    url = "https://pypi.org/pypi/%s/json" % pkg
    try:
        r = requests.get(url)
        return r.json()["info"]["version"]
    except:
        return None



import os

from petrus._core import utils
from petrus._core.calcs.Calc import Calc


class File(Calc):
    def _calc_gitignore(self):
        return ".gitignore"

    def _calc_license(self):
        ans = self.prog.pp["project", "license", "file"]
        if type(ans) is str:
            return ans
        return self._find("LICENSE.txt")

    def _calc_main(self):
        n = self.prog.project.name
        return os.path.join("src", n, "__main__.py")

    def _calc_init(self):
        n = self.prog.project.name
        return os.path.join("src", n, "__init__.py")

    def _calc_manifest(self):
        return "MANIFEST.in"

    def _calc_pp(self):
        return "pyproject.toml"

    def _calc_readme(self):
        ans = self.prog.pp["project", "readme"]
        if type(ans) is str:
            return ans
        return self._find("README.rst")

    def _calc_setup(self):
        return "setup.cfg"

    def exists(self, name):
        f = getattr(self, name)
        return os.path.exists(f)

    @staticmethod
    def _find(file):
        if utils.isfile(file):
            return file
        t = os.path.splitext(file)[0]
        l = os.listdir()
        l = list(l)
        l.sort(reverse=True)
        for x in l:
            if t == os.path.splitext(x)[0]:
                return x
        return file



import tomllib

import tomli_w


class TOML:
    def __init__(self, *, text):
        self._data = tomllib.loads(text)

    def __delitem__(self, index):
        self.__setitem__(index, None)

    def __getitem__(self, index):
        keys = self._keys(index)
        ans = self.clone()._data
        for k in keys:
            try:
                ans = ans[k]
            except KeyError:
                return None
        return ans

    def __setitem__(self, index, value):
        keys = self._keys(index)
        ans = self._data
        while len(keys) > 1:
            if type(ans) is dict and type(keys[1]) is str:
                ans.setdefault(keys[0], {})
            ans = ans[keys.pop(0)]
        if value is None:
            del ans[keys[0]]
            return
        if len(keys) > 0:
            ans[keys[0]] = TOML._purge(value)
        elif type(value) is dict:
            self._data = TOML._purge_dict(value)
        else:
            raise TypeError(value)

    def __str__(self):
        return tomli_w.dumps(self._data)

    @staticmethod
    def _keys(index):
        if type(index) is str:
            return [index]
        else:
            return list(index)

    @staticmethod
    def _purge(value):
        if type(value) is dict:
            return TOML._purge_dict(value)
        if type(value) in (list, tuple):
            return TOML._purge_list(value)
        if type(value) in (bool, float, int, str):
            return value
        raise TypeError

    @staticmethod
    def _purge_dict(value):
        ans = dict()
        for k, v in value.items():
            if v is not None:
                ans[k] = v
        return ans

    @staticmethod
    def _purge_list(value):
        ans = list()
        for x in value:
            if x is not None:
                ans.append(x)
        return ans

    def clone(self):
        return type(self)(text=str(self))



import dataclasses
import typing

import packaging.version

from petrus._core import utils


@dataclasses.dataclass(frozen=True)
class Version:
    epoch: typing.Any
    release: typing.Any
    pre: typing.Any
    post: typing.Any
    dev: typing.Any
    local: typing.Any

    def __add__(self, other):
        if type(other) is not type(self):
            other = type(self).parse(other)
        d = dict()
        for n in "epoch release pre post dev local".split():
            a = getattr(self, n)
            b = getattr(other, n)
            d[n] = self._add(a, b, name=n)
        ans = type(self)(**d)
        return ans

    def __radd__(self, other):
        if type(other) is not type(self):
            other = type(self).parse(other)
        return other + self

    def __str__(self) -> str:
        ans = ""
        if self.epoch != 0:
            ans += f"{self.epoch}!"
        ans += ".".join(str(x) for x in self.release)
        if self.pre is not None:
            ans += "".join(str(x) for x in self.pre)
        if self.post is not None:
            ans += f".post{self.post}"
        if self.dev is not None:
            ans += f".dev{self.dev}"
        if self.local is not None:
            ans += f"+{self.local}"
        return ans

    @staticmethod
    def _add(a, b, *, name):
        if name != "release":
            if {a, b} == {None}:
                return None
            if a is None:
                return b
            if b is None:
                return a
            return a + b
        a = dict(enumerate(a))
        b = dict(enumerate(b))
        l = [0] * max(len(a), len(b))
        for i in range(l):
            l[i] += a.get(i, 0)
            l[i] += b.get(i, 0)
        ans = tuple(l)
        return ans

    def _apply_plus(self, length):
        release = list(self.release)
        release += [0] * 3
        index = 3 - length
        release[index] += 1
        release = release[: index + 1]
        release += [0] * 3
        release = release[:3]
        release = tuple(release)
        ans = dataclasses.replace(self, release=release)
        return ans

    def apply(self, arg, /):
        arg = str(arg)
        if arg == "":
            return dataclasses.replace(self)
        if arg in "+ ++ +++".split():
            return self._apply_plus(len(arg))
        return type(self).parse(arg)

    @classmethod
    def parse(cls, text):
        data = packaging.version.parse(text)
        keys = "epoch release pre post dev local".split()
        kwargs = dict()
        for k in keys:
            kwargs[k] = getattr(data, k)
        ans = cls(**kwargs)
        return ans



import os
import subprocess

from petrus._core.calcs.Calc import Calc


class Git(Calc):
    def __call__(self, *args, force=False):
        for a in args:
            if type(a) is not str:
                raise TypeError(a)
        if not (force or self.is_repo()):
            return
        args = ["git"] + list(args)
        return subprocess.run(args)

    def _calc_author(self):
        a = self.prog.kwargs["author"]
        e = self.prog.kwargs["email"]
        if {a, e} == {None}:
            return None
        if a is None:
            return e
        if e is None:
            return a
        return f"{a} <{e}>"

    def init(self):
        if self.is_repo():
            return
        self("init", os.getcwd(), force=True)
        self.commit("Initial Commit")

    def commit_version(self):
        m = "Version %s" % self.prog.project.version
        self.commit(m)

    def commit(self, message):
        if message is None:
            message = "a"
        else:
            message = str(message)
        try:
            self("add", "-A").check_returncode()
        except:
            return
        args = ["commit", "--allow-empty", "-m", message]
        if self.author is not None:
            args += ["--author", self.author]
        self(*args)

    def push(self): ...
    def is_repo(self):
        called = self("rev-parse", force=True)
        if called is None:
            return False
        return not called.returncode

    def move(self, a, b, /):
        try:
            self("mv", a, b).check_returncode()
        except:
            pass
        else:
            return
        os.rename(a, b)
